
{*******************************************************}
{                                                       }
{       FuturaSoft Visual Component Library             }
{       TRuler, Version 1.04 - 25. November 1997        }
{       For Delphi 1.0, 2.0 and 3.0                     }
{                                                       }
{       Copyright (c) 1996-1997 Martin Krmer           }
{                                                       }
{       Martin Krmer                                   }
{       Adolf-Kolping-Str. 4                            }
{       D49179 Ostercappeln                            }
{       (Germany)                                       }
{                                                       }
{       E-mail: kraemer@mathematik.Uni-Osnabrueck.DE    }
{                                                       }
{*******************************************************}

unit Ruler;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, Menus;

{$IFDEF WIN32}
{$IFNDEF VER90}
{$DEFINE DELPHI3}  { Delphi Version >= 3.0 }
{$ENDIF}
{$ENDIF}

{$DEFINE AUTOSTYLE}

type
  TWindowsStyle = (wsAutoDetect, wsWin31, wsWin40);

const
{$IFDEF AUTOSTYLE}
  WindowsStyle : TWindowsStyle = wsAutoDetect;
{$ELSE}
{$IFDEF WIN32}
  WindowsStyle = wsWin40;
{$ELSE}
  WindowsStyle = wsWin31;
{$ENDIF}
{$ENDIF}

const
  beText = 0;
  beLoMetric = 254;
  beHiMetric = 2540;
  beLoEnglish = 100;
  beHiEnglish = 1000;
  beTWIPS = 1440;

type
  TRulerTabKind = (tkDefault, tkLeftJustify, tkCenter, tkRightJustify,
    tkDecimal);

  TRulerKind = (rbHorizontal, rbVertical);

  TRulerOption = (roShowMousePos, roShowCursorPos, roShowTabMarks,
    roTabEditing);
  TRulerOptions = set of TRulerOption;

  TRulerColumn = record
    Min : Integer;
    Max : Integer;
  end;

  {
                    |-----------------------------------------------|
  Gerteeinheiten:  Left/Top                                        Right/Bottom
                    |
                    |-----------------------|
  "                 0                       PixelOrigin
                                            |
                          |-----------------|
  Log. Einheiten:         0                 BaseOrigin
                          |
                          |-----|-----|-----------------|-----|
  "                       0     Min   ScaleOrigin       Max   Size
                          .           |                       .
                          '--|--'--'--#--'--'--|--'--'--#--'--'
                                      | ||              |
  Log. Skaleneinh.:                   0 |ScaleDivision  ScaleInterval
                                        ScaleUnit

  Mastabsverhltnisse:
  Gerteeinheiten = Log. Einheiten * PixelExtent/BaseExtent
                  = Log. Skaleneinh. * ScalePixelExtent/ScaleExtent
  }

  TRuler = class(TCustomControl)
  private
    { Private-Deklarationen }
    FTrueBaseExtent : Integer;
    FTruePixelExtent : Integer;
    FTrueScaleExtent : Integer;
    FTrueScalePixelExtent : Integer;
    FTextLeading : Integer;
    FTextOLAscent : Integer;
    FRulerBitmap : TBitmap;
    FSavedCursorBitmap : TBitmap;
    FColumns : TList;
    FBaseExtent : Integer;
    FBaseOrigin : Integer;
    FCursorPos : Integer;
    FKind : TRulerKind;
    FMax : Integer;
    FMin : Integer;
    FMousePos : Integer;
    FOptions : TRulerOptions;
    FPixelExtent : Integer;
    FPixelOrigin : Integer;
    FScaleDivision : Integer;
    FScaleExtent : Integer;
    FScaleInterval : Integer;
    FScaleOrigin : Integer;
    FScalePixelExtent : Integer;
    FScaleUnit : Integer;
    FSize : Integer;
    FOnPaint : TNotifyEvent;
    procedure AdjustFontMetrics;
    procedure DrawTabMarks;
    procedure UpdateCursorPos(AShowing : Boolean);
    procedure UpdateMousePos(AShowing : Boolean);
    function GetColumn(Index : Integer) : TRulerColumn;
    function GetColumnCount : Integer;
    procedure SetBaseExtent(Value : Integer);
    procedure SetBaseOrigin(Value : Integer);
    function GetBorderStyle : TBorderStyle;
    procedure SetBorderStyle(Value : TBorderStyle);
    procedure SetCursorPos(Value : Integer);
    procedure SetKind(Value : TRulerKind);
    procedure SetMax(Value : Integer);
    procedure SetMin(Value : Integer);
    procedure SetMousePos(Value : Integer);
    procedure SetOptions(Value : TRulerOptions);
    procedure SetPixelExtent(Value : Integer);
    procedure SetPixelOrigin(Value : Integer);
    procedure SetScaleDivision(Value : Integer);
    procedure SetScaleExtent(Value : Integer);
    procedure SetScaleInterval(Value : Integer);
    procedure SetScaleOrigin(Value : Integer);
    procedure SetScalePixelExtent(Value : Integer);
    procedure SetScaleUnit(Value : Integer);
    procedure SetSize(Value : Integer);
    function GetTransparent : Boolean;
    procedure SetTransparent(Value : Boolean);
    procedure CMCtl3DChanged(var Message : TMessage); message CM_CTL3DCHANGED;
    procedure CMFontChanged(var Message : TMessage); message CM_FONTCHANGED;
    procedure CMColorChanged(var Message : TMessage); message CM_COLORCHANGED;
    procedure WMEraseBkgnd(var Message : TWMEraseBkgnd); message WM_ERASEBKGND;
  protected
    { Protected-Deklarationen }
    procedure CreateParams(var Params : TCreateParams); override;
    procedure Paint; override;
    procedure SetParent(AParent : TWinControl); override;
  public
    { Public-Deklarationen }
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    procedure InsertColumn(AColumn : TRulerColumn);
    procedure RemoveColumn(AColumn : TRulerColumn);
    property Canvas;
    property Columns[Index : Integer] : TRulerColumn read GetColumn;
    property ColumnCount : Integer read GetColumnCount;
    property CursorPos : Integer read FCursorPos write SetCursorPos;
    property MousePos : Integer read FMousePos write SetMousePos;
  published
    { Published-Deklarationen }
    property BaseExtent : Integer read FBaseExtent write SetBaseExtent;
    property BaseOrigin : Integer read FBaseOrigin write SetBaseOrigin
      default 0;
    property BorderStyle : TBorderStyle read GetBorderStyle write SetBorderStyle
      default bsSingle;
    property Kind : TRulerKind read FKind write SetKind default rbHorizontal;
    property Max : Integer read FMax write SetMax default High(Integer);
    property Min : Integer read FMin write SetMin default 0;
    property Options : TRulerOptions read FOptions write SetOptions default [];
    property PixelExtent : Integer read FPixelExtent write SetPixelExtent
      default 0;
    property PixelOrigin : Integer read FPixelOrigin write SetPixelOrigin
      default 0;
    property ScaleDivision : Integer read FScaleDivision write SetScaleDivision;
    property ScaleExtent : Integer read FScaleExtent write SetScaleExtent;
    property ScaleInterval : Integer read FScaleInterval write SetScaleInterval;
    property ScaleOrigin : Integer read FScaleOrigin write SetScaleOrigin
      default 0;
    property ScalePixelExtent : Integer read FScalePixelExtent
      write SetScalePixelExtent default 0;
    property ScaleUnit : Integer read FScaleUnit write SetScaleUnit;
    property Size : Integer read FSize write SetSize default -1;
    property Transparent : Boolean read GetTransparent write SetTransparent
      default False;

    property Align;
    property Color;
    property Ctl3D;
    property DragCursor;
    property DragMode;
    property Enabled;
    property Font;
    property ParentColor default False;
    property ParentCtl3D;
    property ParentFont default False;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
    property OnPaint : TNotifyEvent read FOnPaint write FOnPaint;
  end;

procedure Register;

implementation

{$IFDEF WIN32}
  {$R *32}
{$ELSE}
  {$R *}
{$ENDIF}

{ Implementationsteil von TRuler }

const
  RulerBitmapWidth = 8;
  RulerBitmapHeight = 8;
  RulerBitmapVertHOffset = (Ord(High(TRulerTabKind)) + 1) * RulerBitmapWidth;
  RulerBitmapNormVOffset = 0;
  RulerBitmapGrayVOffset = RulerBitmapHeight;
  RulerBitmapMaskVOffset = 2 * RulerBitmapHeight;
  RulerBitmapHotSpots : array[Low(TRulerTabKind)..High(TRulerTabKind)] of Byte
    = (3, 1, 3, 5, 3);

var
  RulerBitmap : TBitmap;

procedure FreeBitmaps; far;
begin
  if RulerBitmap <> nil then
  begin
    RulerBitmap.Free;
    RulerBitmap := nil;
  end;
end;

procedure ReadBitmaps;
var
  BitmapMask : TBitmap;
begin
  if RulerBitmap = nil then
  begin
{$IFNDEF WIN32}
    AddExitProc(FreeBitmaps);
{$ENDIF}
    RulerBitmap := TBitmap.Create;
    RulerBitmap.Handle := LoadBitmap(HInstance, 'RULERBITMAPS');
    (*BitmapMask := TBitmap.Create;
    try
      BitmapMask.Assign(RulerBitmap);
      { Color which will be transparent }
      BitmapMask.Canvas.Brush.Color := clFuchsia;
      BitmapMask.Monochrome := True;
      BitmapMask.Canvas.Brush.Color := clBlack;
      BitmapMask.Canvas.Font.Color := clWhite;
      BitmapMask.Monochrome := False;
      RulerBitmap.Height := RulerBitmap.Height * 2;
      RulerBitmap.Canvas.CopyMode := cmSrcAnd;
      RulerBitmap.Canvas.Draw(0, 0, BitmapMask);
      RulerBitmap.Canvas.CopyMode := cmNotSrcCopy;
      RulerBitmap.Canvas.Draw(0, BitmapMask.Height, BitmapMask);
    finally
      BitmapMask.Free;
    end;*)
  end;
end;

constructor TRuler.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);
  Width := 121;
  Height := 15;
  ControlStyle := [csCaptureMouse, csClickEvents, csFramed, csOpaque,
    csDoubleClicks];
  with Font do
  begin
    Name := 'Small Fonts';
    Size := 7;
  end;
  ParentColor := False;
  BaseExtent := beLoMetric;
  FCursorPos := -1;
  FMax := High(FMax);
  FMousePos := -1;
  FScaleDivision := 25;
  ScaleExtent := beLoMetric;
  FScaleInterval := 100;
  FScaleUnit := 100;
  FSize := -1;
  ReadBitmaps;
  FRulerBitmap := RulerBitmap;
  FSavedCursorBitmap := TBitmap.Create;
  FSavedCursorBitmap.Width := RulerBitmapWidth;
  FSavedCursorBitmap.Height := RulerBitmapHeight;
end;

destructor TRuler.Destroy;
begin
  if FRulerBitmap <> RulerBitmap then
    FRulerBitmap.Free;
  FSavedCursorBitmap.Free;
  if FColumns <> nil then
    FColumns.Free;
  inherited Destroy;
end;

procedure TRuler.CreateParams(var Params : TCreateParams);
begin
  inherited CreateParams(Params);
  with Params do
  begin
    {if csFramed in ControlStyle then
      Style := Style or WS_BORDER;}
    if not (csOpaque in ControlStyle) then
      ExStyle := ExStyle or WS_EX_TRANSPARENT;
  end;
end;

procedure TRuler.SetParent(AParent : TWinControl);
begin
  inherited SetParent(AParent);
  if AParent <> nil then
    AdjustFontMetrics;
end;

procedure TRuler.CMCtl3DChanged(var Message : TMessage);
begin
  inherited;
  Invalidate;
end;

procedure TRuler.CMFontChanged(var Message : TMessage);
begin
  AdjustFontMetrics;
  inherited;  { Impliziert Aufruf von Invalidate }
end;

procedure TRuler.CMColorChanged(var Message : TMessage);
begin
  if Color = clWhite then
  begin
    if FRulerBitmap <> RulerBitmap then
    begin
      FRulerBitmap.Free;
      FRulerBitmap := RulerBitmap;
    end;
  end
  else begin
    if FRulerBitmap = RulerBitmap then
    begin
      FRulerBitmap := TBitmap.Create;
      FRulerBitmap.Assign(RulerBitmap);
    end;
    with FRulerBitmap do
    begin
      Canvas.Brush.Color := Self.Color;
      Canvas.BrushCopy(Bounds(0, 0, Width, Height div 2),
        RulerBitmap, Bounds(0, 0, Width, Height div 2), clWhite);
    end;
  end;
  inherited;  { Impliziert Aufruf von Invalidate }
end;

procedure TRuler.AdjustFontMetrics;
var
  DC : HDC;
  SaveFont : HFont;
  SysMetrics, Metrics : TTextMetric;
  F : TForm;
  PixelsPerInch : Integer;
begin
  DC := GetDC(0);
  GetTextMetrics(DC, SysMetrics);
  SaveFont := SelectObject(DC, Font.Handle);
  GetTextMetrics(DC, Metrics);
  SelectObject(DC, SaveFont);
  ReleaseDC(0, DC);
  FTextLeading := Metrics.tmInternalLeading;
  FTextOLAscent := Metrics.tmAscent - FTextLeading;
  if (FPixelExtent = 0) or (FScalePixelExtent = 0) then
  begin
    F := TForm(GetParentForm(Self));
    if F = nil then
      PixelsPerInch := Screen.PixelsPerInch
    else
      PixelsPerInch := F.PixelsPerInch;
    if FPixelExtent = 0 then
    begin
      FTruePixelExtent := PixelsPerInch;
      if FBaseExtent = beText then
        FTrueBaseExtent := PixelsPerInch;
    end;
    if FScalePixelExtent = 0 then
    begin
      FTrueScalePixelExtent := PixelsPerInch;
      if FScaleExtent = beText then
        FTrueScaleExtent := PixelsPerInch;
    end;
  end;
end;

procedure TRuler.WMEraseBkgnd(var Message : TWMEraseBkgnd);
begin
  inherited;
  if not (csOpaque in ControlStyle) then
    Paint;
end;

procedure BrushOverlay(ACanvas : TCanvas; const Dest : TRect; Bitmap : TBitmap;
  const Source : TRect; Color : TColor);
const
  ROP_DSPDxax = $00E20746;
var
  SavedColor : TColor;
  crBack, crText : TColorRef;
begin
  if Bitmap = nil then Exit;
  with ACanvas do
  begin
    SavedColor := Brush.Color;
    Brush.Color := Color;
    if Bitmap.Monochrome then
    begin
      crText := SetTextColor(Handle, 0);
      crBack := SetBkColor(Handle, $FFFFFF);
    end;
    StretchBlt(Handle, Dest.Left, Dest.Top, Dest.Right - Dest.Left,
      Dest.Bottom - Dest.Top, Bitmap.Canvas.Handle, Source.Left, Source.Top,
      Source.Right - Source.Left, Source.Bottom - Source.Top, ROP_DSPDxax);
    if Bitmap.Monochrome then
    begin
      SetTextColor(Handle, crText);
      SetBkColor(Handle, crBack);
    end;
    Brush.Color := SavedColor;
  end;
end;

procedure TRuler.Paint;
var
  R : TRect;
  TrueStartX, StartX, MinX, MaxX, EndX, TrueEndX : Integer;
  ScaleOriginX : Integer;
  ScaleRoundDiff, ScaleRoundCorrection : Integer;
  MarkPosition : Longint;
  ScaleIntervalMid : Integer;
  MarkPositionX, MarkTextOffsetX, MarkTextOffsetY : Integer;
  MarkText : string;
  XL : Longint;

  procedure DrawWindow(Left, Top, Right, Bottom : Integer; Color : TColor);
  var
    LTFColor, RBFColor : TColor;
    Framed3D : Boolean;
  begin
    with Canvas do
    begin
      Framed3D := csFramed in ControlStyle;
      if csOpaque in ControlStyle then
        Brush.Color := Color
      else if Framed3D then
        Brush.Style := bsClear
      else
        Exit;
      if Framed3D then
      begin
        Framed3D := Ctl3D;
        if not Framed3D then
        begin
          if Right < R.Right then Inc(Right);
          if Bottom < R.Bottom then Inc(Bottom);
        end;
        if Framed3D and (Color = clBtnFace) then
        begin
          RBFColor := clBtnShadow;
          LTFColor := clBtnHighlight;
        end
        else begin
          RBFColor := clBtnFace;
          LTFColor := clWindowFrame;
        end;
      end
      else
        LTFColor := Color;
      Pen.Color := LTFColor;
      Rectangle(Left, Top, Right, Bottom);
      if Framed3D then
      begin
        Pen.Color := RBFColor;
        Dec(Right);
        Dec(Bottom);
        { Ein guter Compiler sollte die nachfolgende Anweisung
          so wie $IFDEF AUTOSTYLE optimieren. }
        if WindowsStyle >= wsWin40 then
          { Windows 95 Layout }
          Dec(Left);
        PolyLine([Point(Right, Top), Point(Right, Bottom), Point(Left, Bottom)]);
      end;
    end;
  end;

  procedure DrawHorzWindows;
  begin
    with R do
    begin
      if Left < StartX then
        DrawWindow(Left, Top, StartX, Bottom, clBtnShadow)
      else
        StartX := Left;
      if StartX < MinX then
        DrawWindow(StartX, Top, MinX, Bottom, clBtnFace)
      else
        MinX := StartX;
      if EndX < Right then
        DrawWindow(EndX, Top, Right, Bottom, clBtnShadow)
      else
        EndX := Right;
      if MaxX < EndX then
        DrawWindow(MaxX, Top, EndX, Bottom, clBtnFace)
      else
        MaxX := EndX;
      if MinX < MaxX then
        DrawWindow(MinX, Top, MaxX, Bottom, Color);
    end;
  end;

  procedure DrawVertWindows;
  begin
    with R do
    begin
      if Top < StartX then
        DrawWindow(Left, Top, Right, StartX, clBtnShadow)
      else
        StartX := Top;
      if StartX < MinX then
        DrawWindow(Left, StartX, Right, MinX, clBtnFace)
      else
        MinX := StartX;
      if EndX < Bottom then
        DrawWindow(Left, EndX, Right, Bottom, clBtnShadow)
      else
        EndX := Bottom;
      if MaxX < EndX then
        DrawWindow(Left, MaxX, Right, EndX, clBtnFace)
      else
        MaxX := EndX;
      if MinX < MaxX then
        DrawWindow(Left, MinX, Right, MaxX, Color);
    end;
  end;

  procedure DrawHorzMark;
  begin
    with Canvas do
    begin
      if MarkPosition mod FScaleInterval = 0 then
      begin
        MarkText := IntToStr(Abs(MarkPosition div FScaleUnit));
        MarkTextOffsetX := TextWidth(MarkText) div 2;
        if (MarkPositionX - MarkTextOffsetX > StartX)
          and (MarkPositionX + MarkTextOffsetX < EndX) then
        begin
          TextOut(MarkPositionX - MarkTextOffsetX, MarkTextOffsetY, MarkText);
          StartX := MarkPositionX + MarkTextOffsetX;
        end
        else if (MarkPositionX - MarkTextOffsetX > TrueStartX)
          and (MarkPositionX + MarkTextOffsetX < TrueEndX) then
        begin
          TextRect(R, MarkPositionX - MarkTextOffsetX, MarkTextOffsetY,
            MarkText);
          StartX := MarkPositionX + MarkTextOffsetX;
        end;
      end
      else if MarkPosition mod ScaleIntervalMid = 0 then
      begin
        MoveTo(MarkPositionX, 5);
        LineTo(MarkPositionX, 10);
      end
      else
      begin
        MoveTo(MarkPositionX, 6);
        LineTo(MarkPositionX, 8);
      end;
    end;
  end;

  procedure DrawVertMark;
  begin
    with Canvas do
    begin
      if MarkPosition mod FScaleInterval = 0 then
      begin
        MarkText := IntToStr(Abs(MarkPosition div FScaleUnit));
        MarkTextOffsetX := (TextWidth(MarkText) + 1) div 2;
        if (MarkPositionX - MarkTextOffsetY > StartX)
          and (MarkPositionX + MarkTextOffsetY < EndX) then
        begin
          TextOut(7 - MarkTextOffsetX, MarkPositionX - MarkTextOffsetY,
            MarkText);
          StartX := MarkPositionX + MarkTextOffsetY - FTextLeading;
        end
        else if (MarkPositionX - MarkTextOffsetY > TrueStartX)
          and (MarkPositionX + MarkTextOffsetY < TrueEndX) then
        begin
          TextRect(R, 7 - MarkTextOffsetX, MarkPositionX - MarkTextOffsetY,
            MarkText);
          StartX := MarkPositionX + MarkTextOffsetY - FTextLeading;
        end;
      end
      else if MarkPosition mod ScaleIntervalMid = 0 then
      begin
        MoveTo(5, MarkPositionX);
        LineTo(10, MarkPositionX);
      end
      else
      begin
        MoveTo(6, MarkPositionX);
        LineTo(8, MarkPositionX);
      end;
    end;
  end;

begin
  StartX := MulDiv(-FBaseOrigin, FTruePixelExtent, FTrueBaseExtent)
    + FPixelOrigin;
  TrueStartX := StartX;
  if FMin = Low(FMin) then
    MinX := Low(MinX)
  else
    MinX := MulDiv(FMin - FBaseOrigin, FTruePixelExtent, FTrueBaseExtent)
      + FPixelOrigin;
  if FMax = High(FMax) then
    MaxX := High(MaxX)
  else
    MaxX := MulDiv(FMax - FBaseOrigin, FTruePixelExtent, FTrueBaseExtent)
      + FPixelOrigin;
  if (FSize < 0) or (FSize = High(FSize)) then
    EndX := High(EndX)
  else
    EndX := MulDiv(FSize - FBaseOrigin, FTruePixelExtent, FTrueBaseExtent)
      + FPixelOrigin;
  TrueEndX := EndX;
  R := ClientRect;
  Canvas.Pen.Mode := pmCopy;
  if FKind = rbHorizontal then
    DrawHorzWindows
  else
    DrawVertWindows;
  if Ctl3D and (csFramed in ControlStyle) then
  begin
    Dec(EndX);
    Dec(TrueEndX);
  end;
  if FKind = rbHorizontal then
  begin
    R.Left := StartX;
    R.Right := EndX;
  end
  else begin
    R.Top := StartX;
    R.Bottom := EndX;
  end;
  with Canvas do
  begin
    Brush.Style := bsClear;
    Font := Self.Font;
    Pen.Color := Self.Font.Color;
    if FKind = rbHorizontal then
      MarkTextOffsetY := 4 - FTextLeading
    else
      MarkTextOffsetY := FTextLeading + FTextOLAscent div 2;
    ScaleIntervalMid := FScaleInterval div 2;

    ScaleOriginX := MulDiv(FScaleOrigin - FBaseOrigin, FTruePixelExtent,
      FTrueBaseExtent) + FPixelOrigin;
    ScaleRoundDiff := Longint(FScaleOrigin - FBaseOrigin) * FTruePixelExtent
      - Longint(ScaleOriginX - FPixelOrigin) * FTrueBaseExtent;
    ScaleRoundCorrection := MulDiv(ScaleRoundDiff, FTrueScaleExtent,
      FTrueBaseExtent);
    { Erste Teilungsmarke bestimmen }
    MarkPosition := Longint(StartX - ScaleOriginX) * FTrueScaleExtent;
    if MarkPosition >= 0 then
      Inc(MarkPosition, FTrueScalePixelExtent div 2)
    else
      Dec(MarkPosition, FTrueScalePixelExtent div 2);  { Windows: MulDiv ohne +/-1! }
      {Dec(MarkPosition, (FTrueScalePixelExtent + 1) div 2 - 1);}
    MarkPosition := MarkPosition div FTrueScalePixelExtent;
    MarkPosition := (MarkPosition div FScaleDivision) * FScaleDivision;
    {MarkPosition := -FScaleOrigin mod FScaleDivision;
    if MarkPosition <= 0 then
      Inc(MarkPosition, FScaleDivision);
    Inc(MarkPosition, FScaleOrigin);}
    while True do
    begin
      XL := MarkPosition * FTrueScalePixelExtent + ScaleRoundCorrection;
      if XL >= 0 then
        Inc(XL, FTrueScaleExtent div 2)
      else
        Dec(XL, FTrueScaleExtent div 2);  { Windows: MulDiv ohne +/-1! }
        {Dec(XL, (FTrueScaleExtent + 1) div 2 - 1);}
      MarkPositionX := XL div FTrueScaleExtent + ScaleOriginX;
      if MarkPositionX >= EndX then
        Break;
      if MarkPositionX > StartX then
        if FKind = rbHorizontal then
          DrawHorzMark
        else
          DrawVertMark;
      Inc(MarkPosition, FScaleDivision);
    end;
  end;
  if Assigned(FOnPaint) then FOnPaint(Self);
  if roShowTabMarks in FOptions then
    DrawTabMarks;
  UpdateCursorPos(True);
  UpdateMousePos(True);
end;

procedure TRuler.DrawTabMarks;
var
  I : Integer;
begin
end;

procedure TRuler.UpdateCursorPos(AShowing : Boolean);
var
  CursorPosX : Integer;
  CursorPosY : Integer;
  RulerBitmapHOffset : Integer;
  RulerBitmapVOffset : Integer;
begin
  if (roShowCursorPos in FOptions) and (FCursorPos >= 0) then
  begin
    CursorPosX := FPixelOrigin + FCursorPos - RulerBitmapHotSpots[tkDefault];
    if FKind = rbHorizontal then
    begin
      CursorPosY := Height - RulerBitmapHeight + 1;
      RulerBitmapHOffset := 0;
    end
    else begin
      CursorPosY := CursorPosX;
      CursorPosX := Width - RulerBitmapWidth + 1;
      RulerBitmapHOffset := RulerBitmapVertHOffset;
    end;
    RulerBitmapVOffset := RulerBitmapNormVOffset;
    with Canvas do
    begin
      if AShowing then
      begin
        FSavedCursorBitmap.Canvas.CopyRect(Bounds(0, 0, RulerBitmapWidth, RulerBitmapHeight),
          Canvas, Bounds(CursorPosX, CursorPosY, RulerBitmapWidth, RulerBitmapHeight));
        CopyMode := cmSrcAnd;
        CopyRect(Bounds(CursorPosX, CursorPosY, RulerBitmapWidth, RulerBitmapHeight),
          FRulerBitmap.Canvas, Bounds(RulerBitmapHOffset, RulerBitmapMaskVOffset,
          RulerBitmapWidth, RulerBitmapHeight));
        CopyMode := cmSrcPaint;
        CopyRect(Bounds(CursorPosX, CursorPosY, RulerBitmapWidth, RulerBitmapHeight),
          FRulerBitmap.Canvas, Bounds(RulerBitmapHOffset, RulerBitmapVOffset,
          RulerBitmapWidth, RulerBitmapHeight));
        CopyMode := cmSrcCopy;
      end
      else
        CopyRect(Bounds(CursorPosX, CursorPosY, RulerBitmapWidth, RulerBitmapHeight),
          FSavedCursorBitmap.Canvas, Bounds(0, 0, RulerBitmapWidth, RulerBitmapHeight));
    end;
  end;
end;

procedure TRuler.UpdateMousePos(AShowing : Boolean);
begin
  if (roShowMousePos in FOptions) and (FMousePos >= 0) then
    with Canvas do
    begin
      Pen.Color := clBtnFace;
      Pen.Mode := pmNotXor;
      if FKind = rbHorizontal then
      begin
        MoveTo(FMousePos, 1);
        LineTo(FMousePos, Height - 1);
      end
      else begin
        MoveTo(1, FMousePos);
        LineTo(Width - 1, FMousePos);
      end;
    end;
end;

procedure TRuler.InsertColumn(AColumn : TRulerColumn);
begin
  if FColumns = nil then
    FColumns := TList.Create;
  FColumns.Add(@AColumn);
end;

procedure TRuler.RemoveColumn(AColumn : TRulerColumn);
begin
  FColumns.Remove(@AColumn);
  if FColumns.Count = 0 then
  begin
    FColumns.Free;
    FColumns := nil;
  end;
end;

function TRuler.GetColumn(Index : Integer) : TRulerColumn;
begin
  Result := TRulerColumn(FColumns.Items[Index]^);
end;

function TRuler.GetColumnCount : Integer;
begin
  Result := FColumns.Count;
end;

procedure TRuler.SetBaseExtent(Value : Integer);
begin
  if FBaseExtent <> Value then
  begin
    FBaseExtent := Value;
    if Value = beText then
      FTrueBaseExtent := FTruePixelExtent
    else
      FTrueBaseExtent := Value;
    Invalidate;
  end;
end;

procedure TRuler.SetBaseOrigin(Value : Integer);
begin
  if FBaseOrigin <> Value then
  begin
    FBaseOrigin := Value;
    Invalidate;
  end;
end;

function TRuler.GetBorderStyle : TBorderStyle;
begin
  if csFramed in ControlStyle then
    Result := bsSingle
  else
    Result := bsNone;
end;

procedure TRuler.SetBorderStyle(Value : TBorderStyle);
begin
  if BorderStyle <> Value then
  begin
    if Value = bsNone then
      ControlStyle := ControlStyle - [csFramed]
    else
      ControlStyle := ControlStyle + [csFramed];
    {RecreateWnd;}
    Invalidate;
  end;
end;

procedure TRuler.SetCursorPos(Value : Integer);
begin
  if FCursorPos <> Value then
  begin
    UpdateMousePos(False);
    UpdateCursorPos(False);
    FCursorPos := Value;
    UpdateCursorPos(True);
    UpdateMousePos(True);
  end;
end;

procedure TRuler.SetKind(Value : TRulerKind);
begin
  if FKind <> Value then
  begin
    FKind := Value;
    if not (csLoading in ComponentState) then
      SetBounds(Left, Top, Height, Width);
    Invalidate;
  end;
end;

procedure TRuler.SetMax(Value : Integer);
begin
  if FMax <> Value then
  begin
    FMax := Value;
    Invalidate;
  end;
end;

procedure TRuler.SetMin(Value : Integer);
begin
  if FMin <> Value then
  begin
    FMin := Value;
    Invalidate;
  end;
end;

procedure TRuler.SetMousePos(Value : Integer);
begin
  if FMousePos <> Value then
  begin
    UpdateMousePos(False);
    FMousePos := Value;
    UpdateMousePos(True);
  end;
end;

procedure TRuler.SetOptions(Value : TRulerOptions);
const
  LayoutOptions = [roShowMousePos, roShowCursorPos, roShowTabMarks];
var
  ChangedOptions : TRulerOptions;
begin
  if FOptions <> Value then
  begin
    ChangedOptions := (FOptions + Value) - (FOptions * Value);
    FOptions := Value;
    if ChangedOptions * LayoutOptions <> [] then
      Invalidate;
  end;
end;

procedure TRuler.SetPixelExtent(Value : Integer);
begin
  if FPixelExtent <> Value then
  begin
    FPixelExtent := Value;
    if Value = 0 then
      AdjustFontMetrics
    else begin
      FTruePixelExtent := Value;
      if FBaseExtent = beText then
        FTrueBaseExtent := Value;
    end;
    Invalidate;
  end;
end;

procedure TRuler.SetPixelOrigin(Value : Integer);
begin
  if FPixelOrigin <> Value then
  begin
    FPixelOrigin := Value;
    Invalidate;
  end;
end;

procedure TRuler.SetScaleDivision(Value : Integer);
begin
  if (Value > 0) and (FScaleDivision <> Value) then
  begin
    FScaleDivision := Value;
    Invalidate;
  end;
end;

procedure TRuler.SetScaleExtent(Value : Integer);
begin
  if FScaleExtent <> Value then
  begin
    FScaleExtent := Value;
    if Value = beText then
      FTrueScaleExtent := FTrueScalePixelExtent
    else
      FTrueScaleExtent := Value;
    Invalidate;
  end;
end;

procedure TRuler.SetScaleInterval(Value : Integer);
begin
  if (Value > 0) and (FScaleInterval <> Value) then
  begin
    FScaleInterval := Value;
    Invalidate;
  end;
end;

procedure TRuler.SetScaleOrigin(Value : Integer);
begin
  if FScaleOrigin <> Value then
  begin
    FScaleOrigin := Value;
    Invalidate;
  end;
end;

procedure TRuler.SetScalePixelExtent(Value : Integer);
begin
  if FScalePixelExtent <> Value then
  begin
    FScalePixelExtent := Value;
    if Value = 0 then
      AdjustFontMetrics
    else begin
      FTrueScalePixelExtent := Value;
      if FScaleExtent = beText then
        FTrueScaleExtent := Value;
    end;
    Invalidate;
  end;
end;

procedure TRuler.SetScaleUnit(Value : Integer);
begin
  if (Value > 0) and (FScaleUnit <> Value) then
  begin
    FScaleUnit := Value;
    Invalidate;
  end;
end;

procedure TRuler.SetSize(Value : Integer);
begin
  if FSize <> Value then
  begin
    FSize := Value;
    Invalidate;
  end;
end;

function TRuler.GetTransparent : Boolean;
begin
  Result := not (csOpaque in ControlStyle);
end;

procedure TRuler.SetTransparent(Value : Boolean);
begin
  if Transparent <> Value then
  begin
    if Value then
      ControlStyle := ControlStyle - [csOpaque]
    else
      ControlStyle := ControlStyle + [csOpaque];
    RecreateWnd;
  end;
end;

{ Registrierung }

procedure Register;
begin
  RegisterComponents('Samples', [TRuler]);
end;

{$IFDEF AUTOSTYLE}
initialization
  if WindowsStyle = wsAutoDetect then
    if NewStyleControls then
      WindowsStyle := wsWin40
    else
      WindowsStyle := wsWin31;
{$ENDIF}

{$IFDEF WIN32}
finalization
  FreeBitmaps;
{$ENDIF}
end.
